/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <list>

#ifdef USE_EGL
#include "composite-canvas-egl.h"
#elif USE_GLX
#include "composite-canvas-glx.h"
#endif
#include "composite-canvas-pixman.h"
#include "composite-canvas-xrender.h"

#include "composite-test.h"
#include "benchmark.h"
#include "options.h"
#include "log.h"

static const char *default_benchmarks[] = {
    "default",
    "fade",
    "brick",
    "pixman",
    "xrender",
    NULL
};

static void
list_tests()
{
    const map<string, CompositeTest *> &tests = Benchmark::tests();

    for (map<string, CompositeTest *>::const_iterator test_iter = tests.begin();
         test_iter != tests.end();
         test_iter++)
    {
        CompositeTest *test = test_iter->second;
        if (test->name().empty())
            continue;

        Log::info("[Test] %s\n", test->name().c_str());

        const map<string, CompositeTest::Option> &options = test->options();

        for (map<string, CompositeTest::Option>::const_iterator opt_iter = options.begin();
             opt_iter != options.end();
             opt_iter++)
        {
            const CompositeTest::Option &opt = opt_iter->second;
            Log::info("  [Option] %s\n"
                      "    Description  : %s\n"
                      "    Default Value: %s\n",
                      opt.name.c_str(), 
                      opt.description.c_str(),
                      opt.default_value.c_str());
        }
    }
}

static void
add_default_benchmarks(std::list<Benchmark*> &benchmarks)
{
    for (const char **s = default_benchmarks; *s != NULL; s++)
        benchmarks.push_back(new Benchmark(*s));
}

static void
add_custom_benchmarks(std::list<Benchmark *> &benchmarks)
{
    for (std::list<string>::const_iterator iter = Options::benchmarks.begin();
         iter != Options::benchmarks.end();
         iter++)
    {
        Log::debug("Adding benchmark '%s'\n", iter->c_str());
        benchmarks.push_back(new Benchmark(*iter));
    }
}

void
assign_canvas_to_tests(std::list<CompositeCanvas *> &canvas_list)
{
    const map<string, CompositeTest *> &tests = Benchmark::tests();

    for (map<string, CompositeTest *>::const_iterator test_iter = tests.begin();
         test_iter != tests.end();
         test_iter++)

    {
        CompositeTest *test = test_iter->second;
        if (test->type().empty())
            continue;

        for (std::list<CompositeCanvas *>::const_iterator iter = canvas_list.begin();
             iter != canvas_list.end();
             iter++)
        {
            CompositeCanvas *canvas = *iter;
            if (canvas->type() == test->type()) {
                test->set_canvas(canvas);
                break;
            }
        }
    }
}

static void
next_benchmark(std::list<Benchmark*>& benchmarks,
               std::list<Benchmark*>::const_iterator &benchIt)
{
    benchIt++;
    if (benchIt == benchmarks.end() && Options::run_forever)
        benchIt = benchmarks.begin();
}


int
main(int argc, char **argv)
{
    std::list<CompositeCanvas *> canvas_list;
    canvas_list.push_back(new CompositeCanvasPixman());
    canvas_list.push_back(new CompositeCanvasXRender());

#ifdef USE_EGL
    canvas_list.push_back(new CompositeCanvasEGL());
#elif USE_GLX
    canvas_list.push_back(new CompositeCanvasGLX());
#endif

    std::list<Benchmark*> benchmarks;

    if (!Options::parse_args(argc, argv))
        return 1;

    if (Options::show_help) {
        Options::print_help();
        return 0;
    }

    Benchmark::register_test(*new CompositeTestDefaultOptions());
    Benchmark::register_test(*new CompositeTestSimpleDefault());
    Benchmark::register_test(*new CompositeTestSimpleFade());
    Benchmark::register_test(*new CompositeTestSimpleBrick());
    Benchmark::register_test(*new CompositeTestPixman());
    Benchmark::register_test(*new CompositeTestXRender());

    // Assign a compatible canvas to each test
    assign_canvas_to_tests(canvas_list);

    if (Options::list_tests) {
        list_tests();
        return 0;
    }

    // Add the benchmarks to run
    if (Options::benchmarks.empty())
        add_default_benchmarks(benchmarks);
    else
        add_custom_benchmarks(benchmarks);


    Log::info("=======================================================\n");
    Log::info("    glcompbench %s\n", GLCOMPBENCH_VERSION);

    CompositeCanvas *canvas = 0;
    std::list<Benchmark *> current_benchmarks;

    // Find continuous runs of benchmarks that use the same canvas
    // and run them together
    for (std::list<Benchmark *>::const_iterator iter = benchmarks.begin();
         iter != benchmarks.end();
         next_benchmark(benchmarks, iter))
    {
        Benchmark *benchmark = *iter;
        CompositeCanvas *test_canvas = benchmark->get_test().canvas();

        // Check the canvas of the next benchmark (if any)
        std::list<Benchmark *>::const_iterator next_iter = iter;
        next_iter++;
        CompositeCanvas *next_canvas = next_iter == benchmarks.end() ? 0 :
                                       (*next_iter)->get_test().canvas();

        // If we have a change in canvas, deinit the previous one
        // and init the new
        if (test_canvas != canvas && test_canvas != 0) {
            if (canvas)
                canvas->deinit();

            if (!test_canvas->init()) {
                Log::error("Cannot initialize canvas of type '%s'. "
                           "Skipping benchmark based on test '%s'.\n",
                           test_canvas->type().c_str(),
                           benchmark->get_test().name().c_str());
                continue;
            }

            canvas = test_canvas;
            Log::info("=======================================================\n");
            Log::info("%s", canvas->info_string().c_str());
            Log::info("=======================================================\n");
        }

        current_benchmarks.push_back(benchmark);

        // We should run the current benchmarks if we have reached the end
        // or the next benchmark will cause a change to a new valid canvas
        bool should_run = canvas &&
                          (next_iter == benchmarks.end() ||
                           (next_canvas != 0 && next_canvas != canvas));

        if (should_run) {
            canvas->run_tests(current_benchmarks);
            current_benchmarks.clear();
        }

    }

    return 0;
}
